//
//  BasicS3Folder.swift
//  BasicS3
//

import Foundation

class BasicS3Folder : NSObject, XMLParserDelegate {
    fileprivate var parser:XMLParser?
    fileprivate var currentString:String = ""
    fileprivate var currentObject:BasicS3Object?
    fileprivate var inCommonPrefix:Bool = false
    fileprivate var lastModifiedDateFormatter:DateFormatter
    fileprivate (set) var objects:[BasicS3Object] = []
    fileprivate (set) var folders:[BasicS3Folder] = []
    fileprivate var indexedFolders:[String:BasicS3Folder] = [:]

    
    /* The subfolders found on the most recent call to parseXML */
    fileprivate (set) var newSubfolders:[String] = []
    var folderName = ""
    /* Includes trailing / (except at the top-level, where folderPath == "") */
    fileprivate var folderPath = ""
    
    override init () {
        lastModifiedDateFormatter = DateFormatter()
        lastModifiedDateFormatter.locale = Locale(identifier:"en_US_POSIX")
        lastModifiedDateFormatter.dateFormat = "yyyy-MM-dd'T'HH:mm:ss.SSS'Z'"
        lastModifiedDateFormatter.timeZone = TimeZone(secondsFromGMT: 0)
    }
    
    func navigate (relativePath: String) -> BasicS3Folder? {
        if relativePath == "" { return self }
        
        let folderNames = (relativePath.hasSuffix("/") ? String(relativePath[..<relativePath.index(before: relativePath.endIndex)]) : relativePath).components(separatedBy: "/")
        var currentFolder:BasicS3Folder? = self
        
        for f in folderNames {
            currentFolder = currentFolder?.indexedFolders[f]
        }
        
        return currentFolder
    }
    
    func recursiveObjectNames () -> [String] {
        return objects.map() { folderPath + $0.name! } + folders.map() { $0.recursiveObjectNames() }.joined()
    }
    
    func parseXML(xml: Data) -> Bool {
        parser = XMLParser(data: xml)
        defer { parser = nil }
        
        newSubfolders = []
        
        parser!.delegate = self
        return parser!.parse()
    }

    fileprivate struct Elements {
        static let ListBucketResult = "ListBucketResult"
        static let Name = "Name"
        static let Prefix = "Prefix"
        static let KeyCount = "KeyCount"
        static let MaxKeys = "MaxKeys"
        static let Delimiter = "Delimiter"
        static let IsTruncated = "IsTruncated"
        static let Contents = "Contents"
        static let Key = "Key"
        static let LastModified = "LastModified"
        static let ETag = "ETag"
        static let Size = "Size"
        static let StorageClass = "StorageClass"
        static let CommonPrefixes = "CommonPrefixes"
    }

    func parser(_ parser: XMLParser, didStartElement elementName: String, namespaceURI: String?, qualifiedName qName: String?, attributes attributeDict: [String : String] = [:]) {
        switch elementName {
        case Elements.ListBucketResult:
            if attributeDict["xmlns"] != "http://s3.amazonaws.com/doc/2006-03-01/" {
                NSLog("Error parsing S3 object list XML - xmlns unexpectedly \(attributeDict["xmlns"]!)")
                parser.abortParsing()
            }
            break
        case Elements.Name:
            currentString = ""
        case Elements.Prefix:
            currentString = ""
        case Elements.KeyCount:
            currentString = ""
        case Elements.MaxKeys:
            currentString = ""
        case Elements.Delimiter:
            currentString = ""
        case Elements.IsTruncated:
            currentString = ""
        case Elements.Contents:
            currentObject = BasicS3Object()
        case Elements.Key:
            currentString = ""
        case Elements.LastModified:
            currentString = ""
        case Elements.ETag:
            currentString = ""
        case Elements.Size:
            currentString = ""
        case Elements.StorageClass:
            currentString = ""
        case Elements.CommonPrefixes:
            inCommonPrefix = true
        default:
            break
        }
    }

    func parser(_ parser: XMLParser, didEndElement elementName: String, namespaceURI: String?, qualifiedName qName: String?) {
        switch elementName {
        case Elements.ListBucketResult:
            break
        case Elements.Name:
            break
        case Elements.Prefix: /* AWS uses the same tag "Prefix" with two different meanings */
            if inCommonPrefix {
                let newFolder = BasicS3Folder()
                let newFolderName = String(currentString[..<currentString.index(before: currentString.endIndex)]) /* Remove trailing / */
                newFolder.folderName = newFolderName
                folders.append(newFolder)
                newSubfolders.append(newFolderName)
                indexedFolders[newFolderName] = newFolder
            }
            else {
                folderPath = currentString
            }
        case Elements.KeyCount:
            break
        case Elements.MaxKeys:
            break
        case Elements.Delimiter:
            break
        case Elements.IsTruncated:
            break
        case Elements.Contents:
            objects.append(currentObject!)
            currentObject = nil
        case Elements.Key:
            if !currentString.hasPrefix(folderPath) { parser.abortParsing() }
            currentObject!.name = String(currentString[currentString.index(currentString.startIndex, offsetBy: folderPath.distance(from: folderPath.startIndex, to: folderPath.endIndex))...])
        case Elements.LastModified:
            currentObject!.lastModified = lastModifiedDateFormatter.date(from: currentString)
        case Elements.ETag:
            currentObject!.etag = currentString.replacingOccurrences(of: "\"", with: "")
        case Elements.Size:
            currentObject!.size = Int(currentString)
        case Elements.StorageClass:
            break
        case Elements.CommonPrefixes:
            inCommonPrefix = false
        default:
            break
        }
    }

    func parser(_ parser: XMLParser, foundCharacters string: String) {
        currentString += string
    }
}
